pub use private::Blocktree;

mod private {
    use btserde::{read_from, write_to};
    use fuse_backend_rs::{
        abi::fuse_abi::{Attr, CreateIn},
        api::filesystem::{
            Context, Entry, FileSystem, FsOptions, OpenOptions, ZeroCopyReader, ZeroCopyWriter,
        },
    };
    use serde::{Deserialize, Serialize};
    use std::{
        collections::hash_map::{self, HashMap},
        ffi::CStr,
        fs::File,
        io::{self, SeekFrom, Write},
        path::{Path, PathBuf},
        sync::{
            atomic::{AtomicU64, AtomicUsize, Ordering},
            RwLock,
        },
        time::Duration,
    };

    use crate::{
        crypto::Creds, Block, BlockOpenOptions, BoxInIoErr, DirEntry, Directory, Error, Result,
    };

    type Inode = u64;

    #[repr(u64)]
    pub enum SpecInodes {
        Sb = 1,
        RootDir = 2,
        FirstFree = 11,
    }

    impl From<SpecInodes> for Inode {
        fn from(special: SpecInodes) -> Self {
            special as Inode
        }
    }

    type InodeTableValue = (AtomicUsize, Box<dyn Block>);
    type InodeTable = HashMap<Inode, InodeTableValue>;
    type InodeTableEntry<'a> = hash_map::Entry<'a, Inode, InodeTableValue>;

    /// Structure for metadata about a blocktree.
    #[derive(Debug, PartialEq, Eq, Serialize, Deserialize)]
    struct Superblock {
        /// The generation number of the cluster this part of the blocktree is stored on.
        generation: u64,
        /// The next free inode available to the cluster.
        next_inode: u64,
    }

    /// Structure for managing the part of a blocktree which is stored in the local filesystem.
    pub struct Blocktree<C> {
        /// The path to the directory in the local filesystem where this blocktree is located.
        path: PathBuf,
        /// A map from inode numbers to their reference counts.
        inodes: RwLock<InodeTable>,
        /// The next inode that will be assigned to a new block.
        next_inode: AtomicU64,
        /// The generation number of this filesystem. This is the same for every other server in
        /// the same cluster.
        generation: u64,
        /// The credentials this blocktree instance will use for all cryptographic operations.
        creds: C,
    }

    impl<C: Creds + Clone + 'static> Blocktree<C> {
        /// Creates a new empty blocktree at the given path.
        pub fn new_empty(btdir: PathBuf, generation: u64, creds: C) -> Result<Blocktree<C>> {
            // Initialize the superblock.
            let mut sb_block = Self::open_block(&btdir, SpecInodes::Sb.into(), creds.clone())?;
            let sb = Superblock {
                generation,
                next_inode: SpecInodes::FirstFree.into(),
            };
            write_to(&sb, &mut sb_block)?;
            sb_block.flush()?;

            // Initialize the root directory.
            let mut root_block =
                Self::open_block(&btdir, SpecInodes::RootDir.into(), creds.clone())?;
            write_to(&Directory::new(), &mut root_block)?;
            root_block.flush()?;

            Self::new(btdir, sb, sb_block, root_block, creds)
        }

        /// Opens an existing blocktree stored at the given path.
        pub fn new_existing(btdir: PathBuf, creds: C) -> Result<Blocktree<C>> {
            let mut sb_block = Self::open_block(&btdir, SpecInodes::Sb.into(), creds.clone())?;
            let sb = read_from(&mut sb_block)?;
            let root_block = Self::open_block(&btdir, SpecInodes::RootDir.into(), creds.clone())?;
            Self::new(btdir, sb, sb_block, root_block, creds)
        }

        fn new(
            btdir: PathBuf,
            sb: Superblock,
            sb_block: Box<dyn Block>,
            root_block: Box<dyn Block>,
            creds: C,
        ) -> Result<Blocktree<C>> {
            let mut inodes = HashMap::with_capacity(1);
            inodes.insert(SpecInodes::Sb.into(), (AtomicUsize::new(1), sb_block));
            inodes.insert(
                SpecInodes::RootDir.into(),
                (AtomicUsize::new(1), root_block),
            );
            Ok(Blocktree {
                path: btdir,
                inodes: RwLock::new(inodes),
                next_inode: AtomicU64::new(sb.next_inode),
                generation: sb.generation,
                creds,
            })
        }

        /// Returns the path to the file storing the given inode's data.
        fn block_path<P: AsRef<Path>>(parent: P, inode: Inode) -> PathBuf {
            let group = inode / 0xFF;
            let mut path = PathBuf::new();
            path.push(parent);
            path.push(format!("{group:02x}"));
            path.push(format!("{inode:x}.blk"));
            path
        }

        fn open_block<P: AsRef<Path>>(btdir: P, inode: Inode, creds: C) -> Result<Box<dyn Block>> {
            let path = Self::block_path(&btdir, inode);
            let dir = path.ancestors().nth(1).unwrap();
            if let Err(err) = std::fs::create_dir(dir) {
                match err.kind() {
                    io::ErrorKind::AlreadyExists => (),
                    _ => return Err(err.into()),
                }
            }
            let file = std::fs::OpenOptions::new()
                .read(true)
                .write(true)
                .create(true)
                .open(path)?;
            Self::open_block_file(file, creds)
        }

        fn open_block_file(file: File, creds: C) -> Result<Box<dyn Block>> {
            BlockOpenOptions::new()
                .with_creds(creds)
                .with_compress(false)
                .with_encrypt(true)
                .with_inner(file)
                .open()
        }

        fn access_entry<T, F: FnOnce(InodeTableEntry) -> io::Result<T>>(
            &self,
            inode: Inode,
            cb: F,
        ) -> io::Result<T> {
            let mut inodes = self
                .inodes
                .write()
                .map_err(|err| Error::custom(err.to_string()))?;
            let entry = inodes.entry(inode);
            cb(entry)
        }

        fn access_block<T, F: FnOnce(&mut Box<dyn Block>) -> io::Result<T>>(
            &self,
            inode: Inode,
            cb: F,
        ) -> io::Result<T> {
            self.access_entry(inode, |mut entry| match &mut entry {
                InodeTableEntry::Vacant(_) => Err(Error::custom("Inode is not open").into()),
                InodeTableEntry::Occupied(entry) => cb(&mut entry.get_mut().1),
            })
        }

        fn inode_open(&self, inode: Inode) -> io::Result<()> {
            self.access_entry(inode, |entry| match entry {
                InodeTableEntry::Vacant(entry) => {
                    let block = Self::open_block(&self.path, inode, self.creds.clone())?;
                    entry.insert((AtomicUsize::new(1), block));
                    Ok(())
                }
                InodeTableEntry::Occupied(entry) => {
                    entry.get().0.fetch_add(1, Ordering::SeqCst);
                    Ok(())
                }
            })
        }

        fn inode_release(&self, inode: Inode) -> io::Result<()> {
            let prev = {
                let inodes = self
                    .inodes
                    .read()
                    .map_err(|err| Error::custom(err.to_string()))?;
                let pair = inodes.get(&inode).unwrap();
                pair.0.fetch_sub(1, Ordering::SeqCst)
            };
            if 1 == prev {
                let mut inodes = self
                    .inodes
                    .write()
                    .map_err(|err| Error::custom(err.to_string()))?;
                inodes.remove(&inode);
            }
            Ok(())
        }

        fn inode_write(
            &self,
            inode: Inode,
            reader: &mut dyn ZeroCopyReader,
            count: usize,
            off: u64,
        ) -> io::Result<usize> {
            self.access_block(inode, |block| {
                reader.read_to(&mut block.as_mut(), count, off)
            })
        }

        fn inode_read(
            &self,
            inode: Inode,
            writer: &mut dyn ZeroCopyWriter,
            count: usize,
            off: u64,
        ) -> io::Result<usize> {
            self.access_block(inode, |block| {
                writer.write_from(&mut block.as_mut(), count, off)
            })
        }

        fn ser_write<T: Serialize>(&self, inode: Inode, value: &T) -> io::Result<()> {
            self.access_block(inode, |block| {
                block.seek(SeekFrom::Start(0))?;
                write_to(&value, block)?;
                Ok(())
            })
        }

        fn directory_read(&self, inode: Inode) -> io::Result<(Directory, Attr)> {
            self.access_block(inode, |block| {
                block.seek(SeekFrom::Start(0))?;
                let dir = read_from(block)?;
                let attr = block
                    .mut_meta_body()
                    .access_secrets(|secrets| Ok(secrets.attr()))?;
                Ok((dir, attr))
            })
        }

        /// Updates the superblock file with the next valid inode value.
        fn update_sb(&self) -> io::Result<()> {
            let sb = Superblock {
                generation: self.generation,
                next_inode: self.next_inode.load(Ordering::SeqCst),
            };
            self.ser_write(SpecInodes::Sb.into(), &sb)
        }

        fn attr_timeout(&self) -> Duration {
            Duration::from_secs(5)
        }

        fn entry_timeout(&self) -> Duration {
            Duration::from_secs(5)
        }
    }

    impl<C: Creds + Clone + 'static> FileSystem for Blocktree<C> {
        type Inode = Inode;
        type Handle = u64;

        fn init(&self, _capable: FsOptions) -> io::Result<FsOptions> {
            let options = FsOptions::empty()
                .union(FsOptions::ASYNC_READ.complement())
                .union(FsOptions::BIG_WRITES)
                .union(FsOptions::FLOCK_LOCKS)
                .union(FsOptions::WRITEBACK_CACHE)
                .union(FsOptions::POSIX_ACL.complement());
            Ok(options)
        }

        fn lookup(&self, _ctx: &Context, parent: Self::Inode, name: &CStr) -> io::Result<Entry> {
            let (dir, ..) = self.directory_read(parent)?;
            let name = name.to_str().box_err()?;
            let entry = dir.entries.get(name).ok_or_else(|| {
                io::Error::new(
                    io::ErrorKind::NotFound,
                    format!("no entry for {name} found"),
                )
            })?;
            let inode = match entry {
                DirEntry::File(entry) => entry.inode,
                DirEntry::Directory(entry) => entry.inode,
                DirEntry::Server(_) => {
                    return Err(io::Error::new(
                        io::ErrorKind::Unsupported,
                        "can't lookup server entry",
                    ))
                }
            };
            let mut stat =
                self.access_block(inode, |block| Ok(block.meta_body().secrets()?.stat()))?;
            stat.st_ino = inode;
            Ok(Entry {
                inode,
                generation: self.generation,
                attr: stat,
                attr_flags: 0,
                attr_timeout: self.attr_timeout(),
                entry_timeout: self.entry_timeout(),
            })
        }

        fn open(
            &self,
            _ctx: &Context,
            inode: Self::Inode,
            _flags: u32,
            _fuse_flags: u32,
        ) -> io::Result<(Option<Self::Handle>, OpenOptions)> {
            self.inode_open(inode)?;
            Ok((None, OpenOptions::empty()))
        }

        fn create(
            &self,
            _ctx: &Context,
            parent: Self::Inode,
            name: &CStr,
            _args: CreateIn,
        ) -> std::io::Result<(Entry, Option<Self::Handle>, OpenOptions)> {
            // Reserve a free inode.
            // TODO: Obviously this strategy won't work when there are multiples servers in this
            // generation.
            let inode = self.next_inode.fetch_add(1, Ordering::SeqCst);
            self.update_sb()?;

            // Create the new file block.
            self.inode_open(inode)?;

            // Add a directory entry to the parent for the new inode.
            let attr = self.access_block(parent, |dir_block| {
                dir_block.seek(SeekFrom::Start(0))?;
                let mut dir: Directory = read_from(dir_block).box_err()?;
                let name_str = name
                    .to_str()
                    .map_err(|err| io::Error::new(io::ErrorKind::InvalidInput, err))?;
                dir.add_file(name_str.to_owned(), inode)?;
                dir_block.seek(SeekFrom::Start(0))?;
                write_to(&dir, dir_block).box_err()?;
                dir_block.flush()?;
                Ok(dir_block
                    .mut_meta_body()
                    .access_secrets(|secrets| Ok(secrets.attr()))?)
            })?;

            let entry = Entry {
                inode,
                generation: self.generation,
                attr: attr.into(),
                attr_flags: 0,
                attr_timeout: self.attr_timeout(),
                entry_timeout: self.entry_timeout(),
            };
            // TODO: What is OpenOptions used for?
            let options = OpenOptions::empty();
            Ok((entry, None, options))
        }

        fn write(
            &self,
            _ctx: &Context,
            inode: Self::Inode,
            _handle: Self::Handle,
            r: &mut dyn fuse_backend_rs::api::filesystem::ZeroCopyReader,
            size: u32,
            offset: u64,
            _lock_owner: Option<u64>,
            _delayed_write: bool,
            _flags: u32,
            _fuse_flags: u32,
        ) -> io::Result<usize> {
            let count = size.try_into().box_err()?;
            self.inode_write(inode, r, count, offset)
        }

        fn read(
            &self,
            _ctx: &Context,
            inode: Self::Inode,
            _handle: Self::Handle,
            w: &mut dyn fuse_backend_rs::api::filesystem::ZeroCopyWriter,
            size: u32,
            offset: u64,
            _lock_owner: Option<u64>,
            _flags: u32,
        ) -> io::Result<usize> {
            let count = size.try_into().box_err()?;
            self.inode_read(inode, w, count, offset)
        }

        fn release(
            &self,
            _ctx: &Context,
            inode: Self::Inode,
            _flags: u32,
            _handle: Self::Handle,
            flush: bool,
            _flock_release: bool,
            _lock_owner: Option<u64>,
        ) -> io::Result<()> {
            if flush {
                self.access_block(inode, |entry| entry.flush())?;
            };
            self.inode_release(inode)
        }
    }
}

#[cfg(test)]
mod tests {
    use std::ffi::CString;

    use fuse_backend_rs::api::filesystem::FileSystem;
    use tempdir::TempDir;

    use test_helpers::*;

    use crate::test_helpers;

    use super::{private::SpecInodes, *};

    #[test]
    fn create_write_open_read() {
        let dir = TempDir::new("fuse").expect("failed to create temp dir");
        let creds = test_helpers::NODE_CREDS.clone();
        let bt = Blocktree::new_empty(dir.path().to_owned(), 0, creds)
            .expect("failed to create empty blocktree");
        let name = CString::new("README.md").unwrap();
        let (entry, ..) = bt
            .create(
                &Default::default(),
                SpecInodes::RootDir.into(),
                name.as_c_str(),
                Default::default(),
            )
            .expect("failed to create file");
        bt.open(&Default::default(), entry.inode, 0, 0)
            .expect("failed to open file");
        const LEN: usize = 32;
        let mut expected = BtCursor::new([1u8; LEN]);
        let written = bt
            .write(
                &Default::default(),
                entry.inode,
                Default::default(),
                &mut expected,
                LEN as u32,
                0,
                None,
                false,
                0,
                0,
            )
            .expect("write failed");
        assert_eq!(LEN, written);

        let mut actual = BtCursor::new([0u8; LEN]);
        let read = bt
            .read(
                &Default::default(),
                entry.inode,
                Default::default(),
                &mut actual,
                LEN as u32,
                0,
                None,
                0,
            )
            .expect("failed to read");
        assert_eq!(LEN, read);

        assert_eq!(expected, actual)
    }

    #[test]
    fn lookup() {
        let dir = TempDir::new("fuse").expect("failed to create temp dir");
        let creds = test_helpers::NODE_CREDS.clone();
        let bt = Blocktree::new_empty(dir.path().to_owned(), 0, creds)
            .expect("failed to create empty blocktree");
        let name = CString::new("README.md").unwrap();
        let (expected, ..) = bt
            .create(
                &Default::default(),
                SpecInodes::RootDir.into(),
                name.as_c_str(),
                Default::default(),
            )
            .expect("failed to create file");

        let actual = bt
            .lookup(&Default::default(), SpecInodes::RootDir.into(), &name)
            .expect("lookup failed");

        assert_eq!(expected.generation, actual.generation);
        assert_eq!(expected.inode, actual.inode);
        assert_eq!(expected.attr.st_size, actual.attr.st_size);
        assert_eq!(expected.attr.st_uid, actual.attr.st_uid);
        assert_eq!(expected.attr.st_gid, actual.attr.st_gid);
        assert_eq!(expected.attr.st_blksize, actual.attr.st_blksize);
        assert_eq!(expected.attr.st_blocks, actual.attr.st_blocks);
    }
}
