pub use private::Trailered;

mod private {
    use std::{
        io::{self, Read, Seek, SeekFrom, Write},
        marker::PhantomData,
    };

    use btserde::{read_from, write_to};
    use fuse_backend_rs::file_traits::FileReadWriteVolatile;
    use serde::{de::DeserializeOwned, Serialize};

    use crate::{BoxInIoErr, Decompose, MetaAccess, Result, WriteInteg};

    /// A struct which wraps a stream and which writes a trailing data structure to it when flushed.
    pub struct Trailered<T, D> {
        inner: T,
        body_len: u64,
        phantom: PhantomData<D>,
    }

    impl<T: Read + Seek, D: DeserializeOwned> Trailered<T, D> {
        pub fn empty(inner: T) -> Trailered<T, D> {
            Trailered {
                inner,
                body_len: 0,
                phantom: PhantomData,
            }
        }

        /// Creates a new `Trailered<T>` containing the given `T`. This method requires that the given
        /// stream is either empty, or contains a valid serialization of `D` and a the offset at which
        /// `D` is stored.
        pub fn new(mut inner: T) -> Result<(Trailered<T, D>, Option<D>)> {
            let pos = inner.stream_position()?;
            let end = inner.seek(SeekFrom::End(0))?;
            if 0 == end {
                return Ok((Self::empty(inner), None));
            }
            inner.seek(SeekFrom::End(-8))?;
            let offset: i64 = read_from(&mut inner)?;
            let body_len = inner.seek(SeekFrom::End(offset))?;
            let trailer: D = read_from(&mut inner)?;
            inner.seek(SeekFrom::Start(pos))?;
            Ok((
                Trailered {
                    inner,
                    body_len,
                    phantom: PhantomData,
                },
                Some(trailer),
            ))
        }
    }

    impl<T: Seek, D> Trailered<T, D> {
        fn post_write(&mut self, written: usize) -> io::Result<usize> {
            if 0 == written {
                return Ok(0);
            }
            // I cannot return an error at this point because bytes have already been written to inner.
            // So if I can't track the body len due to a failure, a panic is the only option.
            let pos = self
                .inner
                .stream_position()
                .expect("failed to get stream position");
            self.body_len = self.body_len.max(pos);
            Ok(written)
        }
    }

    impl<T: Read + Seek, D> Read for Trailered<T, D> {
        fn read(&mut self, buf: &mut [u8]) -> io::Result<usize> {
            let pos = self.inner.stream_position()?;
            let available_u64 = self.body_len - pos;
            let available: usize = available_u64.try_into().box_err()?;
            let limit = buf.len().min(available);
            self.inner.read(&mut buf[..limit])
        }
    }

    impl<T: Write + Seek, D: Serialize> Trailered<T, D> {
        pub fn write(&mut self, buf: &[u8]) -> io::Result<usize> {
            let written = self.inner.write(buf)?;
            self.post_write(written)
        }

        fn write_trailer(&mut self, trailer: &D) -> io::Result<u64> {
            let pos = self.inner.stream_position()?;
            self.inner.seek(SeekFrom::Start(self.body_len))?;
            write_to(trailer, &mut self.inner).box_err()?;
            let offset_u64 = 8 + self.inner.stream_position()? - self.body_len;
            let offset = -(offset_u64 as i64);
            write_to(&offset, &mut self.inner).box_err()?;
            Ok(pos)
        }

        pub fn flush(&mut self, trailer: &D) -> io::Result<()> {
            let prev_pos = self.write_trailer(trailer)?;
            self.inner.flush()?;
            self.inner.seek(SeekFrom::Start(prev_pos))?;
            Ok(())
        }
    }

    impl<T: WriteInteg + Seek, D: Serialize> Trailered<T, D> {
        pub(crate) fn flush_integ(&mut self, trailer: &D, integrity: &[u8]) -> io::Result<()> {
            let prev_pos = self.write_trailer(trailer)?;
            self.inner.flush_integ(integrity)?;
            self.inner.seek(SeekFrom::Start(prev_pos))?;
            Ok(())
        }
    }

    impl<T: Seek, D> Seek for Trailered<T, D> {
        fn seek(&mut self, pos: SeekFrom) -> io::Result<u64> {
            /// Adds a signed integer to an unsigned integer and returns the result.
            fn add_signed(unsigned: u64, signed: i64) -> u64 {
                if signed >= 0 {
                    unsigned + signed as u64
                } else {
                    unsigned - (-signed as u64)
                }
            }

            let from_start = match pos {
                SeekFrom::Start(from_start) => from_start,
                SeekFrom::Current(from_curr) => {
                    add_signed(self.inner.stream_position()?, from_curr)
                }
                SeekFrom::End(from_end) => add_signed(self.body_len, from_end),
            };
            let from_start = from_start.min(self.body_len);
            self.inner.seek(SeekFrom::Start(from_start))
        }
    }

    impl<T, D> Decompose<T> for Trailered<T, D> {
        fn into_inner(self) -> T {
            self.inner
        }
    }

    impl<T: MetaAccess, D> MetaAccess for Trailered<T, D> {
        fn meta(&self) -> &crate::BlockMeta {
            self.inner.meta()
        }

        fn mut_meta(&mut self) -> &mut crate::BlockMeta {
            self.inner.mut_meta()
        }
    }

    impl<T: FileReadWriteVolatile, D> FileReadWriteVolatile for Trailered<T, D> {
        fn read_volatile(
            &mut self,
            slice: fuse_backend_rs::file_buf::FileVolatileSlice,
        ) -> io::Result<usize> {
            self.inner.read_volatile(slice)
        }

        fn write_volatile(
            &mut self,
            slice: fuse_backend_rs::file_buf::FileVolatileSlice,
        ) -> io::Result<usize> {
            self.inner.write_volatile(slice)
        }

        fn read_at_volatile(
            &mut self,
            slice: fuse_backend_rs::file_buf::FileVolatileSlice,
            offset: u64,
        ) -> io::Result<usize> {
            self.inner.read_at_volatile(slice, offset)
        }

        fn write_at_volatile(
            &mut self,
            slice: fuse_backend_rs::file_buf::FileVolatileSlice,
            offset: u64,
        ) -> io::Result<usize> {
            self.inner.write_at_volatile(slice, offset)
        }
    }
}

#[cfg(test)]
mod tests {
    use crate::Decompose;

    use super::*;
    use std::io::{Cursor, Read, Seek, SeekFrom};

    /// Tests that a new `Trailered<T>` can be created from an empty stream.
    #[test]
    fn trailered_new_empty() {
        let cursor = Cursor::new(Vec::new());

        let (_, trailer): (_, Option<String>) =
            Trailered::new(cursor).expect("Trailered::new failed");

        assert_eq!(None, trailer);
    }

    /// Tests that an error is returned when an attempt is made to create a `Trailered<T>` from a
    /// non-empty stream which is too short.
    #[test]
    fn trailered_new_inner_too_short_is_error() {
        let cursor = Cursor::new([0u8; 5]);

        let result = Trailered::<_, u128>::new(cursor);

        assert!(result.is_err())
    }

    /// Checks that the trailer is persisted to the inner stream.
    #[test]
    fn trailered_trailer_persisted() {
        const EXPECTED: &str = "Everyone deserves to be remembered,";
        let cursor = {
            let cursor = Cursor::new(Vec::new());
            let (mut trailered, trailer) =
                Trailered::<_, String>::new(cursor).expect("Trailered::new failed");
            assert!(trailer.is_none());
            trailered
                .flush(&EXPECTED.to_string())
                .expect("flush failed");
            trailered.into_inner()
        };

        let (_, trailer) = Trailered::<_, String>::new(cursor).expect("Trailered::new failed");

        assert_eq!(EXPECTED, trailer.unwrap());
    }

    #[test]
    fn trailered_written_data_persisted() {
        const EXPECTED: &[u8] = b"and every life has something to teach us.";
        let mut cursor = {
            let (mut trailered, _) = Trailered::<_, u8>::new(Cursor::new(Vec::new()))
                .expect("failed to create first trailered");
            trailered.write(EXPECTED).expect("write failed");
            trailered.flush(&1).expect("flush failed");
            trailered.into_inner()
        };
        cursor.seek(SeekFrom::Start(0)).expect("seek failed");
        let (mut trailered, _) =
            Trailered::<_, u8>::new(cursor).expect("failed to created second trailered");
        let mut actual = vec![0u8; EXPECTED.len()];

        trailered.read(&mut actual).expect("read failed");

        assert_eq!(EXPECTED, actual);
    }

    fn trailered_for_seek_test() -> Trailered<impl Read + Seek, u8> {
        let (mut trailered, _) =
            Trailered::new(Cursor::new(Vec::new())).expect("failed to create trailered");
        trailered
            .write(&[0, 1, 2, 3, 4, 5, 6, 7])
            .expect("write failed");
        trailered.seek(SeekFrom::Start(0)).expect("seek failed");
        trailered
    }

    #[test]
    fn trailered_seek_from_start() {
        const EXPECTED: u8 = 2;
        let mut trailered = trailered_for_seek_test();

        trailered
            .seek(SeekFrom::Start(EXPECTED as u64))
            .expect("seek failed");

        let mut actual = [0u8; 1];
        trailered.read(&mut actual).expect("read failed");
        assert_eq!(EXPECTED, actual[0]);
    }

    #[test]
    fn trailered_seek_from_curr() {
        const EXPECTED: u8 = 5;
        let mut trailered = trailered_for_seek_test();
        trailered
            .seek(SeekFrom::Start(6))
            .expect("seek from start failed");

        trailered
            .seek(SeekFrom::Current(-1))
            .expect("seek from current failed");

        let mut actual = [0u8; 1];
        trailered.read(&mut actual).expect("read failed");
        assert_eq!(EXPECTED, actual[0]);
    }

    #[test]
    fn trailered_seek_from_end() {
        const EXPECTED: u8 = 7;
        let mut trailered = trailered_for_seek_test();

        trailered.seek(SeekFrom::End(-1)).expect("seek failed");

        let mut actual = [0u8; 1];
        trailered.read(&mut actual).expect("read failed");
        assert_eq!(EXPECTED, actual[0]);
    }

    /// Tests that a read past the end of the body in a `Trailered<T>` is not allowed.
    #[test]
    fn trailered_read_limited_to_body_len() {
        let (mut trailered, trailer) =
            Trailered::new(Cursor::new(Vec::new())).expect("failed to create Trailered");
        assert!(trailer.is_none());
        const EXPECTED: &[u8] = &[1, 1, 1, 1, 1, 0, 0, 0];
        trailered.write(&[1u8; 5]).expect("write failed");
        trailered.flush(&1u8).expect("flush failed");
        trailered.seek(SeekFrom::Start(0)).expect("seek failed");
        let mut actual = vec![0u8; EXPECTED.len()];

        // If read goes past the end of the body then there will be a 1 in the sixth position of
        // actual.
        trailered.read(&mut actual).expect("read failed");

        assert_eq!(EXPECTED, actual);
    }
}
